/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ImgCreator.h"

#include <windows.h>
#include <string>
#include <fstream>
#include <sstream>
#include "Logger.hpp"
#include "DeviceReader.h"
#include "util.h"

using namespace std;

ImgCreator::ImgCreator()
{
    LOG_DEBUG("ImgCreator::ImgCreator()");
	mSuccessful = true;
}

string ImgCreator::getError()
{
    LOG_DEBUG("ImgCreator::getError(). Return value: "+mError+".");
	return mError;
}

bool ImgCreator::createImg(string driveLetter, string fileNameOut)
{
    LOG_DEBUG("Entering ImgCreator::createImg()...");
	if(driveLetter.length() == 2 and driveLetter.substr(1,1) == ":")
	{
	    LOG_DEBUG("Removing ':' from drive letter.");
		driveLetter = driveLetter.substr(0,1);
	}
	else if(driveLetter.length() != 1)
	{
		mError = "Invalid drive argument: '"+str(driveLetter)+"'";
		LOG_ERROR(mError);
        LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: false.");
		return false;
	}
    LOG_INFO("Drive letter: " + driveLetter);
	
    LOG_DEBUG("Creating DeviceReader instance.");
	DeviceReader device(driveLetter);
	if(device.failed())
	{
		mError = device.getError();
		LOG_ERROR("Failed to create DeviceReader.");
        LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: false.");
		return false;
	}
    LOG_DEBUG("DeviceReader instance successfully created.");
	
    LOG_DEBUG("Detecting disk size...");
	mTotalBytes = getDiskSize(driveLetter);
	//progress_set_total(mTotalBytes);
    LOG_INFO("CD size (bytes): "+str(mTotalBytes));
    LOG_INFO("CD size (MiB): "+str(static_cast<double>(mTotalBytes)/1048576.0));
    if(mTotalBytes <= 0) {
        LOG_ERROR("No CD found!");
        LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: false.");
        return false;
    }
	
	LOG_DEBUG("Opening output file: "+fileNameOut);
	mOutFile.open(fileNameOut.c_str(), ios::out | ios::binary | ios::trunc);
	if(mOutFile.fail())
	{
		mError = "Failed to open " + fileNameOut;
		LOG_ERROR(mError);
        LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: false.");
		return false;
	}
	
	LOG_INFO("Beginning to read from the CD...");
    int buflen;
    next_progress = 0.0;
	long long bytesWritten = 0;
	bool failed = false;
	while(true)
	{
		buflen = mTotalBytes - bytesWritten;
        if (buflen > 65536) buflen = 65536;
        string data = device.read(buflen);
		if(device.failed()) {
			mError = device.getError();
			LOG_ERROR("Failed to read from device!");
			LOG_DEBUG("buflen = "+str(buflen));
			LOG_DEBUG("bytesWritten = "+str(bytesWritten));
			failed = true;
			break;
		}
		if(data.length() == 0) {
		    LOG_DEBUG("device.read() returned an empty string. Exiting main loop.");
		    break;
		}
		
		mOutFile << data;
		if(mOutFile.bad())
		{
			mError = "Write operation failed on file: " + fileNameOut;
			LOG_ERROR(mError);
			LOG_DEBUG("buflen = "+str(buflen));
			LOG_DEBUG("bytesWritten = "+str(bytesWritten));
			LOG_DEBUG("data.length() = "+str(data.length()));
			failed = true;
			break;
		}
		bytesWritten += data.length();
		updateProgress(bytesWritten);
		if (bytesWritten == mTotalBytes) {
		    LOG_DEBUG("bytesWritten == mTotalBytes. Exiting main loop.");
		    break;
		}
	}
	LOG_DEBUG("Closing output file.");
	mOutFile.close();
	
	if(!failed && bytesWritten != mTotalBytes)
	{
		mError = "The number of bytes written doesn't equal the disk size!";
		LOG_ERROR(mError);
		failed = true;
	}
	
	if(failed)
	{
	    LOG_ERROR("ISO creation failed! Deleting output file...");
		DeleteFile(fileNameOut.c_str());
        LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: false.");
		return false;
	}
    LOG_DEBUG("Returning from ImgCreator::createImg(). Return value: true.");
	return true;
}

void ImgCreator::updateProgress(long long bytesWritten)
{
	ostringstream status;
	status.precision(1);
	double percent = static_cast<double>(bytesWritten)*100.0 / static_cast<double>(mTotalBytes);
	if(percent >= next_progress) {
        status << fixed << static_cast<double>(bytesWritten) / 1048576.0 << " MB (" << percent << "%) of " << static_cast<double>(mTotalBytes) / 1048576.0 << " MiB";
        LOG_INFO(status.str());
        next_progress += 2.5;
	}
}

